theory Syntax imports Main begin

section \<open>Old parsing\<close>
text \<open>Old parsing in ISAC used theories\<close>
ML \<open>
val thy = @{theory};
fun parse thy str = Syntax.read_term_global thy str;
val env = 
  [(parse thy "equ::bool", parse thy "x + (1::int) = 2"),
   (parse thy "var::int",  parse thy "x::int"),
   (parse thy "sol::bool", parse thy "(x::int) = 1")
  ];
val precond = parse thy "(0::int) < x"
\<close>
text \<open>However, user input could not be parsed ...\<close>
ML \<open>parse thy "x = 2 - 1"\<close>
text \<open>... because the type 'int' is lost (we have type 'a now).
  The learner would need to input x::int all the time.\<close>

section \<open>Contexts\<close>
ML \<open>

structure Syntax_Data = Proof_Data
  (type T = term list
   fun init _ = []);

fun update trm = Syntax_Data.map (fn trms => trm::trms);
val ctxt =
  Proof_Context.init_global @{theory} |>
  Syntax_Data.map (fn trms => @{term "x::int"}::trms);

fun parseNEW ctxt str =
  let
    val result = Syntax.read_term ctxt str;
  in
    update result;
    result
  end;
val notJustIt = Syntax_Data.get ctxt;
\<close>

section \<open>New parsing\<close>
text \<open>New parsing in ISAC uses contexts, which are rigorously updated
  by the Lucas-Interpreter. Then ...\<close>
ML \<open>parseNEW ctxt "x"\<close>
text \<open>... x gets type 'int' from the context.\<close>
end
